import datetime
from typing import Any, Dict, List, Type, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.decision import Decision
from ..types import UNSET, Unset

T = TypeVar("T", bound="SupervisionResult")


@_attrs_define
class SupervisionResult:
    """
    Attributes:
        supervision_request_id (UUID):
        created_at (datetime.datetime):
        decision (Decision):
        reasoning (str):
        id (Union[Unset, UUID]):
        chosen_toolrequest_id (Union[Unset, UUID]):
    """

    supervision_request_id: UUID
    created_at: datetime.datetime
    decision: Decision
    reasoning: str
    id: Union[Unset, UUID] = UNSET
    chosen_toolrequest_id: Union[Unset, UUID] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        supervision_request_id = str(self.supervision_request_id)

        created_at = self.created_at.isoformat()

        decision = self.decision.value

        reasoning = self.reasoning

        id: Union[Unset, str] = UNSET
        if not isinstance(self.id, Unset):
            id = str(self.id)

        chosen_toolrequest_id: Union[Unset, str] = UNSET
        if not isinstance(self.chosen_toolrequest_id, Unset):
            chosen_toolrequest_id = str(self.chosen_toolrequest_id)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "supervision_request_id": supervision_request_id,
                "created_at": created_at,
                "decision": decision,
                "reasoning": reasoning,
            }
        )
        if id is not UNSET:
            field_dict["id"] = id
        if chosen_toolrequest_id is not UNSET:
            field_dict["chosen_toolrequest_id"] = chosen_toolrequest_id

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        supervision_request_id = UUID(d.pop("supervision_request_id"))

        created_at = isoparse(d.pop("created_at"))

        decision = Decision(d.pop("decision"))

        reasoning = d.pop("reasoning")

        _id = d.pop("id", UNSET)
        id: Union[Unset, UUID]
        if isinstance(_id, Unset):
            id = UNSET
        else:
            id = UUID(_id)

        _chosen_toolrequest_id = d.pop("chosen_toolrequest_id", UNSET)
        chosen_toolrequest_id: Union[Unset, UUID]
        if isinstance(_chosen_toolrequest_id, Unset):
            chosen_toolrequest_id = UNSET
        else:
            chosen_toolrequest_id = UUID(_chosen_toolrequest_id)

        supervision_result = cls(
            supervision_request_id=supervision_request_id,
            created_at=created_at,
            decision=decision,
            reasoning=reasoning,
            id=id,
            chosen_toolrequest_id=chosen_toolrequest_id,
        )

        supervision_result.additional_properties = d
        return supervision_result

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.hub_stats_assigned_reviews import HubStatsAssignedReviews
    from ..models.hub_stats_review_distribution import HubStatsReviewDistribution


T = TypeVar("T", bound="HubStats")


@_attrs_define
class HubStats:
    """
    Attributes:
        connected_clients (int):
        free_clients (int):
        busy_clients (int):
        assigned_reviews (HubStatsAssignedReviews):
        review_distribution (HubStatsReviewDistribution):
        completed_reviews_count (int):
        pending_reviews_count (int):
        assigned_reviews_count (int):
    """

    connected_clients: int
    free_clients: int
    busy_clients: int
    assigned_reviews: "HubStatsAssignedReviews"
    review_distribution: "HubStatsReviewDistribution"
    completed_reviews_count: int
    pending_reviews_count: int
    assigned_reviews_count: int
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        connected_clients = self.connected_clients

        free_clients = self.free_clients

        busy_clients = self.busy_clients

        assigned_reviews = self.assigned_reviews.to_dict()

        review_distribution = self.review_distribution.to_dict()

        completed_reviews_count = self.completed_reviews_count

        pending_reviews_count = self.pending_reviews_count

        assigned_reviews_count = self.assigned_reviews_count

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "connected_clients": connected_clients,
                "free_clients": free_clients,
                "busy_clients": busy_clients,
                "assigned_reviews": assigned_reviews,
                "review_distribution": review_distribution,
                "completed_reviews_count": completed_reviews_count,
                "pending_reviews_count": pending_reviews_count,
                "assigned_reviews_count": assigned_reviews_count,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.hub_stats_assigned_reviews import HubStatsAssignedReviews
        from ..models.hub_stats_review_distribution import HubStatsReviewDistribution

        d = src_dict.copy()
        connected_clients = d.pop("connected_clients")

        free_clients = d.pop("free_clients")

        busy_clients = d.pop("busy_clients")

        assigned_reviews = HubStatsAssignedReviews.from_dict(d.pop("assigned_reviews"))

        review_distribution = HubStatsReviewDistribution.from_dict(d.pop("review_distribution"))

        completed_reviews_count = d.pop("completed_reviews_count")

        pending_reviews_count = d.pop("pending_reviews_count")

        assigned_reviews_count = d.pop("assigned_reviews_count")

        hub_stats = cls(
            connected_clients=connected_clients,
            free_clients=free_clients,
            busy_clients=busy_clients,
            assigned_reviews=assigned_reviews,
            review_distribution=review_distribution,
            completed_reviews_count=completed_reviews_count,
            pending_reviews_count=pending_reviews_count,
            assigned_reviews_count=assigned_reviews_count,
        )

        hub_stats.additional_properties = d
        return hub_stats

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.supervision_request import SupervisionRequest
    from ..models.supervision_result import SupervisionResult
    from ..models.supervision_status import SupervisionStatus


T = TypeVar("T", bound="SupervisionRequestState")


@_attrs_define
class SupervisionRequestState:
    """
    Attributes:
        supervision_request (SupervisionRequest):
        status (SupervisionStatus):
        result (Union[Unset, SupervisionResult]):
    """

    supervision_request: "SupervisionRequest"
    status: "SupervisionStatus"
    result: Union[Unset, "SupervisionResult"] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        supervision_request = self.supervision_request.to_dict()

        status = self.status.to_dict()

        result: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.result, Unset):
            result = self.result.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "supervision_request": supervision_request,
                "status": status,
            }
        )
        if result is not UNSET:
            field_dict["result"] = result

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.supervision_request import SupervisionRequest
        from ..models.supervision_result import SupervisionResult
        from ..models.supervision_status import SupervisionStatus

        d = src_dict.copy()
        supervision_request = SupervisionRequest.from_dict(d.pop("supervision_request"))

        status = SupervisionStatus.from_dict(d.pop("status"))

        _result = d.pop("result", UNSET)
        result: Union[Unset, SupervisionResult]
        if isinstance(_result, Unset):
            result = UNSET
        else:
            result = SupervisionResult.from_dict(_result)

        supervision_request_state = cls(
            supervision_request=supervision_request,
            status=status,
            result=result,
        )

        supervision_request_state.additional_properties = d
        return supervision_request_state

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Callable, Dict, List, Optional, Literal
from entropy_labs.sentinel_api_client.sentinel_api_client.types import UNSET
from enum import Enum
import random
import json
from pydantic import BaseModel, Field
from entropy_labs.mocking.policies import MockPolicy
from threading import Lock
from inspect_ai.solver import TaskState
from inspect_ai.tool import ToolCall, Tool
from inspect_ai.model import ChatMessage, ChatMessageAssistant, ChatMessageTool, ModelOutput, ChatMessageUser, ChatMessageSystem
from uuid import UUID, uuid4
from inspect_ai._util.content import Content, ContentText
from entropy_labs.supervision.langchain.utils import extract_messages_from_events
from entropy_labs.sentinel_api_client.sentinel_api_client.models.task_state import TaskState as APITaskState
from entropy_labs.sentinel_api_client.sentinel_api_client.models.message import Message
from entropy_labs.sentinel_api_client.sentinel_api_client.models.message_role import MessageRole
from entropy_labs.sentinel_api_client.sentinel_api_client.models.message_type import MessageType
from entropy_labs.sentinel_api_client.sentinel_api_client.models.tool_call import ToolCall as ApiToolCall
from entropy_labs.sentinel_api_client.sentinel_api_client.models.tool_call_arguments import ToolCallArguments
from entropy_labs.sentinel_api_client.sentinel_api_client.models.tool import Tool as ApiTool
from entropy_labs.sentinel_api_client.sentinel_api_client.models.output import Output as ApiOutput
import json
import copy
import logging

PREFERRED_LLM_MODEL = "gpt-4o"
DEFAULT_RUN_NAME = "default"




class SupervisionDecisionType(str, Enum):
    APPROVE = "approve"
    REJECT = "reject"
    ESCALATE = "escalate"
    TERMINATE = "terminate"
    MODIFY = "modify"

class ModifiedData(BaseModel):
    tool_args: Optional[List[Any]] = None
    """Modified positional arguments for the tool/function."""

    tool_kwargs: Optional[Dict[str, Any]] = None
    """Modified keyword arguments for the tool/function."""
    
    original_inspect_ai_call: Optional[ToolCall] = None
    """Original InspectAI call that was modified."""
    
    #TODO: Update to support changing the tool call itself

class SupervisionDecision(BaseModel):
    decision: SupervisionDecisionType
    """Supervision decision."""

    modified: Optional[ModifiedData] = Field(default=None)
    """Modified data for decision 'modify'."""

    explanation: Optional[str] = Field(default=None)
    """Explanation for decision."""

class SupervisionContext:
    """
    Context for supervision decisions. This is used to store the context of the currently active project including all tasks/runs.
    """
    def __init__(self, pending_functions: Optional[Dict[str, Dict[str, Any]]] = None):
        self.langchain_events: List[dict] = []  # List to store all logged events
        self.lock = Lock()  # Ensure thread safety
        self.metadata: Dict[str, Any] = {}
        self.inspect_ai_state: Optional[TaskState] = None
        self.openai_messages: List[Dict[str, Any]] = []
        self.anthropic_messages: List[Dict[str, Any]] = []
        self.supervised_functions_registry: Dict[str, Dict[str, Any]] = pending_functions or {}
        self.registered_supervisors: Dict[str, UUID] = {}
        self.local_supervisors_by_id: Dict[UUID, Callable] = {}

    def add_event(self, event: dict):
        with self.lock:
            self.langchain_events.append(event)

    def add_metadata(self, key: str, value: Any):
        self.metadata[key] = value

    def to_text(self) -> str:
        """Converts the supervision context into a textual description."""
        texts = []
        with self.lock:
            # Process LangChain events if any
            if self.langchain_events:
                texts.append("## LangChain Events:")
                for event in self.langchain_events:
                    event_description = self._describe_event(event)
                    texts.append(event_description)

            # Process inspect_ai_state if it exists
            if self.inspect_ai_state:
                inspect_ai_text = self._describe_inspect_ai_state()
                texts.append(inspect_ai_text)
                
            # Process OpenAI messages if any
            if self.openai_messages:
                openai_text = self._describe_openai_messages()
                texts.append(openai_text)

        return "\n\n".join(texts)

    def _describe_event(self, event: dict) -> str:
        """Converts a single event into a textual description."""
        event_type = event.get("event", "Unknown Event")
        data = event.get("data", {})
        description = f"### Event: {event_type}\nData:\n```json\n{json.dumps(data, indent=2)}\n```"
        return description

    def _describe_inspect_ai_state(self) -> str:
        """Converts the inspect_ai_state into a textual description."""
        state = self.inspect_ai_state
        texts = []
        
        if state is None:
            return ""

        # Include meta information
        meta_info = f"## Inspect AI State:\n**Model:** {state.model}\n**Sample ID:** {state.sample_id}\n**Epoch:** {state.epoch}"
        texts.append(meta_info)

        # Include messages
        texts.append("### Messages:")
        for message in state.messages:
            message_text = self._describe_chat_message(message)
            texts.append(message_text)

        # Include output if available
        if state.output:
            texts.append("### Output:")
            texts.append(f"```json\n{json.dumps(state.output.dict(), indent=2)}\n```")

        return "\n\n".join(texts)

    def _describe_chat_message(self, message: ChatMessage) -> str:
        """Converts a chat message into a textual description."""
        role = message.role.capitalize()
        text_content = message.text.strip()
        text = f"**{role}:**\n{text_content}"

        if isinstance(message, ChatMessageAssistant) and message.tool_calls:
            text += "\n\n**Tool Calls:**"
            for tool_call in message.tool_calls:
                tool_call_description = self._describe_tool_call(tool_call)
                text += f"\n{tool_call_description}"

        return text

    def _describe_tool_call(self, tool_call: ToolCall) -> str:
        """Converts a ToolCall into a textual description."""
        description = (
            f"- **Tool Call ID:** {tool_call.id}\n"
            f"  - **Function:** {tool_call.function}\n"
            f"  - **Arguments:** `{json.dumps(tool_call.arguments, indent=2)}`\n"
            f"  - **Type:** {tool_call.type}"
        )
        return description

    def _describe_openai_messages(self) -> str:
        """Converts the OpenAI messages into a textual description."""
        messages = [convert_openai_message(msg) for msg in self.openai_messages]
        texts = []
        texts.append("## OpenAI Messages:")
        for message in messages:
            message_text = self._describe_chat_message(message)
            texts.append(message_text)
        return "\n\n".join(texts)

    # Methods to manage the supervised functions registry
    def add_supervised_function(
        self,
        func: Callable,
        supervision_functions: Optional[List[List[Callable]]] = None,
        ignored_attributes: Optional[List[str]] = None,
    ):
        func_name = func.__qualname__
        with self.lock:
            if func_name in self.supervised_functions_registry:
                print(f"Function '{func_name}' is already registered in context. Skipping.")
                return  # Skip adding the duplicate

            self.supervised_functions_registry[func_name] = {
                'supervision_functions': supervision_functions or [],
                'ignored_attributes': ignored_attributes or [],
                'function': func,
            }
            print(f"Registered function '{func_name}' in supervision context")

    def get_supervised_function_entry(self, func_name: str) -> Optional[Dict[str, Any]]:
        with self.lock:
            return self.supervised_functions_registry.get(func_name)

    def update_tool_id(self, func: Callable, tool_id: UUID):
        with self.lock:
            if func.__qualname__ in self.supervised_functions_registry:
                self.supervised_functions_registry[func.__qualname__]['tool_id'] = tool_id
                print(f"Updated tool ID for '{func.__qualname__}' to {tool_id}")

    def add_run_id_to_supervised_function(self, func: Callable, run_id: UUID):
        with self.lock:
            if func.__qualname__ in self.supervised_functions_registry:
                self.supervised_functions_registry[func.__qualname__]['run_id'] = run_id
                print(f"Updated run ID for '{func.__qualname__}' to {run_id}")

    def add_supervisor_id(self, supervisor_name: str, supervisor_id: UUID):
        with self.lock:
            self.registered_supervisors[supervisor_name] = supervisor_id
            print(f"Locally registered supervisor '{supervisor_name}' with ID: {supervisor_id}")

    def get_supervisor_id(self, supervisor_name: str) -> Optional[UUID]:
        with self.lock:
            return self.registered_supervisors.get(supervisor_name)

    def to_task_state(self) -> APITaskState:
        """Converts the supervision context into the API client's TaskState model."""
        if self.inspect_ai_state:
            return convert_task_state(self.inspect_ai_state)
        elif self.openai_messages: #TODO: this is when user uses directly openai messages, update
            # Convert OpenAI messages to TaskState
            messages = [convert_openai_message(msg) for msg in self.openai_messages]
            empty_output = ApiOutput(
                    model=UNSET,
                    choices=UNSET,
                    usage=UNSET
    )
            return APITaskState(
                    messages=messages,
                    tools=[],
                    output=empty_output,
                    completed=False,
                    tool_choice=UNSET,
                )
        
        elif self.anthropic_messages:
            empty_output = ApiOutput(
                    model=UNSET,
                    choices=UNSET,
                    usage=UNSET
    )
            return APITaskState(
                messages=[convert_anthropic_message(msg) for msg in self.anthropic_messages],
                tools=[],
                output=empty_output,
                completed=False,
                tool_choice=UNSET,
            )
        
        elif self.langchain_events:
            messages = extract_messages_from_events(self.langchain_events)

            # Create your custom TaskState object
            # TODO: Initialize direct API TaskState ?
            task_state = TaskState(
                model='model_name',  # TODO: Update with actual model name
                sample_id=str(uuid4()),
                epoch=1,
                input=messages[0].content if messages else "",
                messages=messages,
                completed=False,
                metadata={}
            )
            return convert_task_state(task_state)
        else:
            # Handle case when no messages are available
            return APITaskState(
                messages=[],
                tools=[],
                output=None,
                completed=False,
                tool_choice=None,
            )
        
    def get_api_messages(self) -> List[Message]:
        """Get the messages from the supervision context in the API client's Message model."""
        if self.openai_messages:
            # Convert OpenAI messages to ChatMessage objects
            return [convert_openai_message(msg) for msg in self.openai_messages]
        elif self.anthropic_messages:
            return [convert_anthropic_message(msg) for msg in self.anthropic_messages]
        elif self.langchain_events:
            # Extract messages from LangChain events
            return [convert_message(msg) for msg in extract_messages_from_events(self.langchain_events)]
        elif self.inspect_ai_state:
            # Use messages from inspect_ai_state
            return [convert_message(msg) for msg in self.inspect_ai_state.messages]
        else:
            # No messages available
            return []
        
    def update_messages(self, messages: List[Dict[str, Any]], anthropic: bool = False):
        """Updates the context with a list of OpenAI messages."""
        with self.lock:
            if anthropic:
                self.anthropic_messages = messages.copy()
            else:
                self.openai_messages = messages.copy()

    def add_local_supervisor(self, supervisor_id: UUID, supervisor_func: Callable, supervisor_name: str):
        """Add a supervisor function to the config."""
        self.local_supervisors_by_id[supervisor_id] = supervisor_func

    def get_supervisor_by_id(self, supervisor_id: UUID) -> Optional[Callable]:
        """Retrieve a supervisor function by its ID."""
        return self.local_supervisors_by_id.get(supervisor_id)

class Run(BaseModel):
    run_id: UUID
    run_name: str
    supervision_context: SupervisionContext = Field(default_factory=SupervisionContext)

    class Config:
        arbitrary_types_allowed = True

class Task(BaseModel):
    task_id: UUID
    task_name: str
    runs: Dict[str, Run] = Field(default_factory=dict)


class Project(BaseModel):
    project_id: UUID
    project_name: str
    tasks: Dict[str, Task] = Field(default_factory=dict)

class SupervisionConfig:
    def __init__(self):
        self.global_supervision_functions: List[Callable] = []
        self.global_mock_policy = MockPolicy.NO_MOCK
        self.override_local_policy = False
        self.mock_responses: Dict[str, List[Any]] = {}
        self.previous_calls: Dict[str, List[Any]] = {}
        self.llm = None
        self.client = None  # Sentinel API client

        # Hierarchical projects structure
        self.projects: Dict[str, Project] = {}  # Mapping from project_name to Project
        self.projects_by_id: Dict[UUID, Project] = {}
        self.tasks_by_id: Dict[UUID, Task] = {}
        self.runs_by_id: Dict[UUID, Run] = {}
        self.runs_by_name: Dict[str, List[Run]] = {}  # New mapping for runs by name
        self.lock = Lock()  # For thread safety
        self.pending_supervised_functions: Dict[str, Dict[str, Any]] = {}

    def set_global_supervision_functions(self, functions: List[Callable]):
        self.global_supervision_functions = functions

    def set_llm(self, llm):
        self.llm = llm

    def set_mock_policy(self, mock_policy: MockPolicy):
        self.global_mock_policy = mock_policy

    def load_previous_execution_log(self, log_file_path: str, log_format='langchain'):
        """Load and process a previous execution log."""
        with open(log_file_path, 'r') as f:
            log_data = f.readlines()
        
        if log_format == 'langchain':
            for line in log_data:
                try:
                    entry = json.loads(line)
                    if entry['event'] == 'on_tool_end':
                        function_name = entry['data']['kwargs'].get('name')
                        output = entry['data']['output']
                        if function_name:
                            if function_name not in self.previous_calls:
                                self.previous_calls[function_name] = []
                            self.previous_calls[function_name].append(output)
                except json.JSONDecodeError:
                    continue  # Skip lines that are not valid JSON
        else:
            raise ValueError(f"Unsupported log format: {log_format}")
        
        # Update mock_responses with examples from previous calls
        self.mock_responses = self.previous_calls.copy()

    def get_mock_response(self, function_name: str) -> Any:
        """Get a mock response for a specific function."""
        if function_name in self.mock_responses:
            return random.choice(self.mock_responses[function_name])
        else:
            raise ValueError(f"No mock responses available for function: {function_name}")

    # Project methods
    def add_project(self, project_name: str, project_id: UUID):
        """Add a new project."""
        project = Project(project_id=project_id, project_name=project_name)
        self.projects[project_name] = project
        self.projects_by_id[project_id] = project  # Add to ID-based dict

    def get_project(self, project_name: str) -> Optional[Project]:
        """Retrieve a project by its name."""
        return self.projects.get(project_name)

    def get_project_by_id(self, project_id: UUID) -> Optional[Project]:
        """Retrieve a project by its ID."""
        return self.projects_by_id.get(project_id)

    # Task methods
    def add_task(self, project_name: str, task_name: str, task_id: UUID):
        """Add a new task to a project."""
        project = self.get_project(project_name)
        if not project:
            raise ValueError(f"Project '{project_name}' does not exist.")
        task = Task(task_id=task_id, task_name=task_name)
        project.tasks[task_name] = task
        self.tasks_by_id[task_id] = task  # Add to ID-based dict

    def get_task(self, project_name: str, task_name: str) -> Optional[Task]:
        """Retrieve a task by its name under a project."""
        project = self.get_project(project_name)
        if project:
            return project.tasks.get(task_name)
        return None

    def get_task_by_id(self, task_id: UUID) -> Optional[Task]:
        """Retrieve a task by its ID."""
        return self.tasks_by_id.get(task_id)

    # Run methods
    def add_run(self, project_name: str, task_name: str, run_name: str, run_id: UUID):
        """Add a new run to a task under a project."""
        with self.lock:
            task = self.get_task(project_name, task_name)
            if not task:
                raise ValueError(f"Task '{task_name}' does not exist under project '{project_name}'.")

            # Obtain a copy of the pending functions
            supervision_context = SupervisionContext(pending_functions=self.pending_supervised_functions)
            print("Initialized supervised functions registry with pending functions.")

            # Optionally, clear the pending supervised functions
            # self.clear_pending_supervised_functions()

            run = Run(run_id=run_id, run_name=run_name, supervision_context=supervision_context)
            task.runs[run_name] = run
            self.runs_by_id[run_id] = run

            # Add to name-based dict
            if run_name not in self.runs_by_name:
                self.runs_by_name[run_name] = []
            self.runs_by_name[run_name].append(run)

    def get_run(self, project_name: str, task_name: str, run_name: str) -> Optional[Run]:
        """Retrieve a run by its name under a task and project."""
        task = self.get_task(project_name, task_name)
        if task:
            return task.runs.get(run_name)
        return None

    def get_run_by_id(self, run_id: UUID) -> Optional[Run]:
        """Retrieve a run by its ID."""
        return self.runs_by_id.get(run_id)

    def get_run_by_name(self, run_name: str) -> Optional[Run]:
        """Retrieve a run by its name across all projects and tasks."""
        for run in self.runs_by_name.get(run_name, []):
            return run
        return None
    
    def get_all_runs(self) -> List[Run]:
        """Retrieve all runs across all projects and tasks."""
        return list(self.runs_by_id.values())

    # Optional editing methods
    def update_project_id(self, project_name: str, new_project_id: UUID):
        """Update the project ID for a given project name."""
        project = self.projects.get(project_name)
        if project:
            project.project_id = new_project_id
        else:
            raise ValueError(f"Project '{project_name}' does not exist.")

    def update_task_id(self, project_name: str, task_name: str, new_task_id: UUID):
        """Update the task ID for a given task name under a project."""
        project = self.projects.get(project_name)
        if project:
            task = project.tasks.get(task_name)
            if task:
                task.task_id = new_task_id
            else:
                raise ValueError(f"Task '{task_name}' does not exist under project '{project_name}'.")
        else:
            raise ValueError(f"Project '{project_name}' does not exist.")

    def update_run_id(self, project_name: str, task_name: str, run_name: str, new_run_id: UUID):
        """Update the run ID for a given run name under a task and project."""
        project = self.projects.get(project_name)
        if project:
            task = project.tasks.get(task_name)
            if task:
                if run_name in task.runs:
                    task.runs[run_name].run_id = new_run_id
                else:
                    raise ValueError(f"Run '{run_name}' does not exist under task '{task_name}' and project '{project_name}'.")
            else:
                raise ValueError(f"Task '{task_name}' does not exist under project '{project_name}'.")
        else:
            raise ValueError(f"Project '{project_name}' does not exist.")

    def update_supervision_context_by_run_id(self, run_id: UUID, new_supervision_context: SupervisionContext):
        """Update the supervision context for a run specified by run_id."""
        run = self.get_run_by_id(run_id)
        if run:
            run.supervision_context = new_supervision_context
            print(f"Updated supervision context for run_id {run_id}")
        else:
            raise ValueError(f"No run found with run_id {run_id}")

    def update_supervision_context_by_run_name(
        self,
        run_name: str,
        new_supervision_context: SupervisionContext,
    ):
        """Update the supervision context for a run specified by run_name."""
        with self.lock:
            runs = self.runs_by_name.get(run_name)
            if runs:
                if len(runs) == 1:
                    run = runs[0]
                else:
                    raise ValueError(
                        f"Multiple runs found with run_name '{run_name}'. Please specify project and task names."
                    )
                run.supervision_context = new_supervision_context
                print(f"Updated supervision context for run_name '{run_name}'")
            else:
                raise ValueError(f"No run found with run_name '{run_name}'")

    def add_supervised_function_to_all_runs(
        self,
        func: Callable,
        supervision_functions: List[List[Callable]] = [],
        ignored_attributes: List[str] = [],
    ):
        """
        Add supervised functions to the supervision context of each run in the supervision configuration.
        """
        with self.lock:
            for run in self.get_all_runs():
                run.supervision_context.add_supervised_function(func, supervision_functions, ignored_attributes)
                print(f"Added supervised function '{func.__qualname__}' to run '{run.run_name}'")
    # Method to register supervised functions temporarily
    def register_pending_supervised_function(
        self,
        func: Callable,
        supervision_functions: Optional[List[List[Callable]]] = None,
        ignored_attributes: Optional[List[str]] = None,
    ):
        func_name = func.__qualname__
        with self.lock:
            if func_name in self.pending_supervised_functions:
                print(f"Function '{func_name}' is already pending registration. Skipping.")
                return  # Skip adding the duplicate

            self.pending_supervised_functions[func_name] = {
                'supervision_functions': supervision_functions or [],
                'ignored_attributes': ignored_attributes or [],
                'function': func,
            }
            print(f"Registered pending supervised function '{func_name}'")

    def get_pending_supervised_functions(self) -> Dict[str, Dict[str, Any]]:
        """Returns a deep copy of the pending supervised functions."""
        with self.lock:
            return copy.deepcopy(self.pending_supervised_functions)

    # Optionally, clear the pending supervised functions if needed
    def clear_pending_supervised_functions(self):
        with self.lock:
            self.pending_supervised_functions.clear()
            print("Cleared pending supervised functions")

# Global instance of SupervisionConfig
supervision_config = SupervisionConfig()

def get_supervision_config():
    return supervision_config

def get_supervision_context(run_id: UUID, project_name: Optional[str] = None, task_name: Optional[str] = None, run_name: Optional[str] = None) -> SupervisionContext:
    if project_name and task_name and run_name:
        run = supervision_config.get_run(project_name, task_name, run_name)
    else:
        run = supervision_config.get_run_by_id(run_id)
    if run:
        return run.supervision_context
    else:
        raise ValueError(f"No run found with run_id {run_id}")

def set_global_supervision_functions(functions: List[Callable]):
    supervision_config.set_global_supervision_functions(functions)

def set_global_mock_policy(mock_policy: MockPolicy, override_local_policy: bool = False):
    supervision_config.set_mock_policy(mock_policy)
    supervision_config.override_local_policy = override_local_policy

def setup_sample_from_previous_calls(log_file_path: str):
    """Set up the SAMPLE_FROM_PREVIOUS_CALLS mock policy."""
    supervision_config.load_previous_execution_log(log_file_path)
    set_global_mock_policy(MockPolicy.SAMPLE_PREVIOUS_CALLS, override_local_policy=True)

def convert_task_state(task_state: TaskState) -> APITaskState:
    """
    Converts Inspect AI TaskState object into the Sentinel API client's TaskState model.
    They should be identical when serialized.
    """
    from entropy_labs.sentinel_api_client.sentinel_api_client.types import UNSET
    from entropy_labs.sentinel_api_client.sentinel_api_client.models.task_state import TaskState as APITaskState

    # Convert messages
    messages = [convert_message(msg) for msg in task_state.messages]

    # Convert tools - we can't do this because Inspect AI doesn't have tool names    
    tools: list[ApiTool] = []

    # Convert output
    output = convert_output(task_state.output)

    # Convert tool_choice
    tool_choice = UNSET

    return APITaskState(
        messages=messages,
        tools=tools,
        output=output,
        completed=False,
        tool_choice=tool_choice,
    )

def convert_message(msg: ChatMessage) -> Message:
    """
    Converts a ChatMessage to a Message in the API model.
    """

    # Convert content to a string
    if isinstance(msg.content, str):
        content_str = msg.content
    elif isinstance(msg.content, list):
        content_str = "\n".join([convert_content(content) for content in msg.content])
    else:
        content_str = ""
        
    if isinstance(msg, ChatMessageTool):
        tool_call_id = msg.tool_call_id
        function = msg.function
        content_str = f"Function {function} has been executed with result: {msg.content}."
        if msg.error:
            content_str = f"Function {function} has been executed with error: {msg.error}"
    else:
        tool_call_id = UNSET
        function = UNSET

    if isinstance(msg, ChatMessageAssistant) and msg.tool_calls:
        tool_calls = [convert_tool_call(tool_call) for tool_call in msg.tool_calls]
    else:
        tool_calls = UNSET

    return Message( 
        source=msg.source if msg.source else UNSET,
        role=MessageRole(msg.role) if msg.role != 'tool' else MessageRole.ASSISTANT,
        content=content_str,
        tool_calls=tool_calls,
        type=MessageType.TEXT #TODO: Add support for audio messages and images
    )

def convert_anthropic_message(msg: Dict) -> Message:
    """
    Converts an Anthropic message dict to a Message object in the API model.
    """

    from entropy_labs.sentinel_api_client.sentinel_api_client.models.message import Message
    from entropy_labs.sentinel_api_client.sentinel_api_client.models.message_type import MessageType
    from entropy_labs.sentinel_api_client.sentinel_api_client.models.message_role import MessageRole
    from entropy_labs.sentinel_api_client.sentinel_api_client.types import UNSET

    # Extract role
    role_str = msg.get('role', 'assistant')
    role_mapping = {
        'assistant': MessageRole.ASSISTANT,
        'user': MessageRole.USER,
        'system': MessageRole.SYSTEM,
    }
    role = role_mapping.get(role_str, MessageRole.ASSISTANT)

    # Extract content
    content = msg.get('content', '')

    content_str = ''
    tool_calls = []
    message_type = MessageType.TEXT  # Default to text

    if isinstance(content, str):
        # Content is a simple string
        content_str = content
    elif isinstance(content, list):
        # Content is a list of content blocks
        for content_block in content:
            block_type = content_block.get('type', 'text')
            if block_type == 'image':
                # Process image content blocks
                image_source = content_block.get('source', {})
                image_data = image_source.get('data', '')
                media_type = image_source.get('media_type', '')
                # Format content as data URI
                content_str = f"data:{media_type};base64,{image_data}"
                message_type = MessageType.IMAGE
                break  # Assume only one image per message
            elif block_type == 'text':
                text = content_block.get('text', '')
                content_str += text + '\n'
            elif block_type == 'tool_use':
                # Process as a tool use
                tool_use = content_block
                api_tool_call = convert_anthropic_tool_call(tool_use)
                tool_calls.append(api_tool_call)
            else:
                # Handle other content block types if needed; for now, we skip
                pass
        else:
            # No image found; set message type to text
            content_str = content_str.strip()
            message_type = MessageType.TEXT
    else:
        # Content is neither string nor list; convert to string
        content_str = str(content)

    # Construct the Message object
    message = Message(
        role=role,
        content=content_str,
        type=message_type,
        tool_calls=tool_calls if tool_calls else UNSET,
    )
    return message


def convert_anthropic_tool_call(tool_use: Dict) -> ApiToolCall:
    """
    Converts an Anthropic tool use dict to an ApiToolCall object.
    """
    from entropy_labs.sentinel_api_client.sentinel_api_client.models.tool_call import ToolCall as ApiToolCall
    from entropy_labs.sentinel_api_client.sentinel_api_client.models.tool_call_arguments import ToolCallArguments
    from entropy_labs.sentinel_api_client.sentinel_api_client.types import UNSET

    id_ = tool_use.get('id', '')
    name = tool_use.get('name', '')
    input_args = tool_use.get('input', {})
    type_ = 'tool_use'  # Define the type as 'tool_use'
    parse_error = UNSET  # No parse error in this context

    # Convert input_args to ToolCallArguments
    arguments_obj = ToolCallArguments.from_dict(input_args)

    tool_call = ApiToolCall(
        id=id_ if id_ else UNSET,
        function=name,
        arguments=arguments_obj,
        type=type_,
        parse_error=parse_error,
    )

    return tool_call


def convert_content(content: Content) -> str:
    # Convert Content to a string representation
    if isinstance(content, ContentText):
        return content.text
    elif hasattr(content, 'data_url'):
        return content.data_url  # For images or other content types
    else:
        return str(content)

def convert_tool_call(tool_call: ToolCall) -> ApiToolCall:
    return ApiToolCall(
        id=tool_call.id,
        function=tool_call.function,
        arguments=ToolCallArguments.from_dict(tool_call.arguments),
        type=tool_call.type
    )


def convert_output(output: ModelOutput) -> ApiOutput:
    from entropy_labs.sentinel_api_client.sentinel_api_client.models.output import Output as ApiOutput

    return ApiOutput(
        model=output.model if output.model else UNSET,
        choices=UNSET, #TODO: Implement if needed
        usage=UNSET
    )

def convert_openai_message(openai_msg: Dict[str, Any]) -> Message:
    """
    Converts an OpenAI message dict to a Message object in the API model.
    """
    from entropy_labs.sentinel_api_client.sentinel_api_client.models.message import Message
    from entropy_labs.sentinel_api_client.sentinel_api_client.models.message_role import MessageRole
    from entropy_labs.sentinel_api_client.sentinel_api_client.models.message_type import MessageType
    from entropy_labs.sentinel_api_client.sentinel_api_client.types import UNSET

    role_str = openai_msg.get('role')
    content = openai_msg.get('content', '') or ''
    message_id = openai_msg.get('id', UNSET)
    source = openai_msg.get('source', UNSET)
    tool_calls_data = openai_msg.get('tool_calls', UNSET)

    # Map role string to MessageRole enum
    role_mapping = {
        'assistant': MessageRole.ASSISTANT,
        'user': MessageRole.USER,
        'system': MessageRole.SYSTEM,
        # 'tool': MessageRole.TOOL,
    }

    role = role_mapping.get(role_str, MessageRole.USER)

    # Determine message type and content
    message_type = MessageType.TEXT  # Default to TEXT
    if isinstance(content, list):
        if len(content) > 1:
            logging.warning("Message has multiple items in content")
        for item in content:
            if item.get('type') == 'image_url':
                message_type = MessageType.IMAGE_URL
                content = item['image_url']['url']
                if content.startswith("data"):
                    message_type = MessageType.IMAGE
                content = item['image_url']['url']
                break
            elif item.get('type') == 'text':
                content = item['text']

    # Convert tool_calls if any
    if tool_calls_data and tool_calls_data is not UNSET:
        tool_calls = [convert_openai_tool_call(tc) for tc in tool_calls_data]
    else:
        tool_calls = UNSET

    message = Message(
        role=role,
        content=content,
        id=message_id,
        type=message_type,
        source=source,
        tool_calls=tool_calls,
    )

    return message

def convert_openai_tool_call(tool_call_dict: Dict[str, Any]) -> ApiToolCall:
    """
    Converts an OpenAI tool call dict to a ToolCall object in the API model.
    """

    id_ = tool_call_dict.get('id', '')
    function = tool_call_dict.get('function', '')
    arguments = tool_call_dict.get('arguments', {})
    type_ = tool_call_dict.get('type', '')
    parse_error = tool_call_dict.get('parse_error', UNSET)
    
    # Ensure 'function' is a string
    if isinstance(function, dict):
        function = function.get('name', '')
    elif not isinstance(function, str):
        function = str(function)

    # Convert arguments to ToolCallArguments
    if isinstance(arguments, dict):
        arguments_obj = ToolCallArguments.from_dict(arguments)
    elif isinstance(arguments, str):
        try:
            arguments_obj = ToolCallArguments.from_dict(json.loads(arguments))
        except json.JSONDecodeError:
            arguments_obj = ToolCallArguments()
    else:
        arguments_obj = ToolCallArguments()

    tool_call = ApiToolCall(
        id=id_,
        function=function,
        arguments=arguments_obj,
        type=type_,
        parse_error=parse_error,
    )

    return tool_call




from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.chain_execution import ChainExecution
    from ..models.supervision_request_state import SupervisionRequestState
    from ..models.supervisor_chain import SupervisorChain


T = TypeVar("T", bound="ChainExecutionState")


@_attrs_define
class ChainExecutionState:
    """
    Attributes:
        chain (SupervisorChain):
        chain_execution (ChainExecution):
        supervision_requests (List['SupervisionRequestState']):
    """

    chain: "SupervisorChain"
    chain_execution: "ChainExecution"
    supervision_requests: List["SupervisionRequestState"]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        chain = self.chain.to_dict()

        chain_execution = self.chain_execution.to_dict()

        supervision_requests = []
        for supervision_requests_item_data in self.supervision_requests:
            supervision_requests_item = supervision_requests_item_data.to_dict()
            supervision_requests.append(supervision_requests_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "chain": chain,
                "chain_execution": chain_execution,
                "supervision_requests": supervision_requests,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.chain_execution import ChainExecution
        from ..models.supervision_request_state import SupervisionRequestState
        from ..models.supervisor_chain import SupervisorChain

        d = src_dict.copy()
        chain = SupervisorChain.from_dict(d.pop("chain"))

        chain_execution = ChainExecution.from_dict(d.pop("chain_execution"))

        supervision_requests = []
        _supervision_requests = d.pop("supervision_requests")
        for supervision_requests_item_data in _supervision_requests:
            supervision_requests_item = SupervisionRequestState.from_dict(supervision_requests_item_data)

            supervision_requests.append(supervision_requests_item)

        chain_execution_state = cls(
            chain=chain,
            chain_execution=chain_execution,
            supervision_requests=supervision_requests,
        )

        chain_execution_state.additional_properties = d
        return chain_execution_state

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Contains all the data models used in inputs/outputs"""

from .arguments import Arguments
from .chain_execution import ChainExecution
from .chain_execution_state import ChainExecutionState
from .chain_request import ChainRequest
from .choice import Choice
from .create_project_body import CreateProjectBody
from .create_run_tool_body import CreateRunToolBody
from .create_run_tool_body_attributes import CreateRunToolBodyAttributes
from .create_task_body import CreateTaskBody
from .decision import Decision
from .error_response import ErrorResponse
from .hub_stats import HubStats
from .hub_stats_assigned_reviews import HubStatsAssignedReviews
from .hub_stats_review_distribution import HubStatsReviewDistribution
from .message import Message
from .message_role import MessageRole
from .message_type import MessageType
from .output import Output
from .project import Project
from .review_payload import ReviewPayload
from .run import Run
from .run_execution import RunExecution
from .status import Status
from .supervision_request import SupervisionRequest
from .supervision_request_state import SupervisionRequestState
from .supervision_result import SupervisionResult
from .supervision_status import SupervisionStatus
from .supervisor import Supervisor
from .supervisor_attributes import SupervisorAttributes
from .supervisor_chain import SupervisorChain
from .supervisor_type import SupervisorType
from .task import Task
from .task_state import TaskState
from .task_state_metadata import TaskStateMetadata
from .task_state_store import TaskStateStore
from .tool import Tool
from .tool_attributes import ToolAttributes
from .tool_call import ToolCall
from .tool_call_arguments import ToolCallArguments
from .tool_choice import ToolChoice
from .tool_request import ToolRequest
from .tool_request_group import ToolRequestGroup
from .update_run_result_body import UpdateRunResultBody
from .usage import Usage

__all__ = (
    "Arguments",
    "ChainExecution",
    "ChainExecutionState",
    "ChainRequest",
    "Choice",
    "CreateProjectBody",
    "CreateRunToolBody",
    "CreateRunToolBodyAttributes",
    "CreateTaskBody",
    "Decision",
    "ErrorResponse",
    "HubStats",
    "HubStatsAssignedReviews",
    "HubStatsReviewDistribution",
    "Message",
    "MessageRole",
    "MessageType",
    "Output",
    "Project",
    "ReviewPayload",
    "Run",
    "RunExecution",
    "Status",
    "SupervisionRequest",
    "SupervisionRequestState",
    "SupervisionResult",
    "SupervisionStatus",
    "Supervisor",
    "SupervisorAttributes",
    "SupervisorChain",
    "SupervisorType",
    "Task",
    "TaskState",
    "TaskStateMetadata",
    "TaskStateStore",
    "Tool",
    "ToolAttributes",
    "ToolCall",
    "ToolCallArguments",
    "ToolChoice",
    "ToolRequest",
    "ToolRequestGroup",
    "UpdateRunResultBody",
    "Usage",
)



from .sentinel_api_client_helper import (
    register_project,
    create_run,
    register_task,
    register_tools_and_supervisors,
    submit_run_status,
    submit_run_result,
    update_run_status,
    update_run_result,
    get_run,
    update_run_status_by_sample_id,
    get_sample_result,
)
from entropy_labs.sentinel_api_client.sentinel_api_client.models.status import Status
from entropy_labs.api.inspect_helpers import (
    register_inspect_samples_with_entropy_labs_solver,
    entropy_labs_web_ui_scorer
)

__all__ = [
    "register_project",
    "create_run",
    "register_task",
    "register_tools_and_supervisors",
    "submit_run_status",
    "submit_run_result",
    "Status",
    "register_inspect_samples_with_entropy_labs_solver",
    "entropy_labs_web_ui_scorer",
    "update_run_status",
    "update_run_result",
    "get_run",
    "update_run_status_by_sample_id",
    "get_sample_result",
]



from typing import Any, get_origin, get_args, Callable, Dict, Any, Union, Optional
import random
import string
import inspect


def create_random_value(return_type: type) -> Any:
    origin = get_origin(return_type)
    args = get_args(return_type)

    if origin is None:
        if return_type == int:
            return random.randint(-1000, 1000)
        elif return_type == float:
            return random.uniform(-1000.0, 1000.0)
        elif return_type == str:
            return ''.join(random.choices(string.ascii_letters + string.digits, k=10))
        elif return_type == bool:
            return random.choice([True, False])
        else:
            raise ValueError(f"Unsupported simple type: {return_type}")
    elif origin is list:
        return [create_random_value(args[0]) for _ in range(random.randint(1, 5))]
    elif origin is dict:
        key_type, value_type = args
        return {create_random_value(key_type): create_random_value(value_type) for _ in range(random.randint(1, 5))}
    elif origin is Union:
        return create_random_value(random.choice(args))
    elif origin is Optional:
        return random.choice([None, create_random_value(args[0])])
    else:
        raise ValueError(f"Unsupported complex type: {return_type}")

def get_function_code(func: Callable) -> str:
    """Retrieve the source code of a function."""
    try:
        return inspect.getsource(func)
    except Exception as e:
        return f"Error retrieving source code: {str(e)}"
    
    



from enum import Enum

class MockPolicy(Enum):
    NO_MOCK = "no_mock"
    SAMPLE_LIST = "sample_from_list" # Sample from a list of responses
    SAMPLE_RANDOM = "sample_random" # Sample a random value from a specified type
    SAMPLE_LLM = "sample_llm" # Sample from an LLM with validation of output format
    SAMPLE_PREVIOUS_CALLS = "sample_from_previous_calls" # Sample from previous calls
    


from typing import Callable
from inspect_ai._util.registry import registry_find
from inspect_ai.solver import TaskState, solver, Solver
from inspect_ai.tool import ToolCall
from inspect_ai import Task
from inspect_ai.dataset import Sample
from entropy_labs.api.sentinel_api_client_helper import (
    register_project, create_run, register_task, register_tools_and_supervisors, 
    update_run_status_by_sample_id, get_sample_result
)
from uuid import UUID
import logging
import yaml
import fnmatch
from inspect_ai.scorer import (
    Score,
    Scorer,
    Target,
    accuracy,
    scorer,
    stderr,
)
from typing import Optional


def _extract_tool_approvers(approve_func: Callable) -> Callable:
    """
    Extract the tool approvers from the Inspect approve function.
    """
    # Get the closure cells of the approve function
    if approve_func.__closure__ is None:
        raise ValueError("No closure found")
    
    # We know tool_approvers is in the first closure cell
    tool_approvers = approve_func.__closure__[0].cell_contents
    
    # Verify we got a callable
    assert callable(tool_approvers), "Expected tool_approvers to be callable"
    # Verify it has the correct name
    assert tool_approvers.__name__ == 'tool_approvers', "Expected function named 'tool_approvers'"
    # Verify it has the expected return type annotation
    assert 'Generator' in str(tool_approvers.__annotations__.get('return', '')), "Expected tool_approvers to return a Generator"
    
    return tool_approvers

def _register_inspect_approvals(run_id: UUID, tool_to_approvers: dict):
    """
    Registers the approvals and tools for the run using the provided tool-to-approvers mapping.

    Args:
        run_id (UUID): The ID of the run.
        tool_to_approvers (dict): A dictionary mapping tools to their respective approvers.
    """
    # Retrieve the run and its supervision context
    from entropy_labs.supervision.config import supervision_config
    run = supervision_config.get_run_by_id(run_id)
    if run is None:
        raise Exception(f"Run with ID {run_id} not found in supervision config.")
    supervision_context = run.supervision_context

    # Iterate over the tool-to-approvers mapping
    for tool_func, approvers in tool_to_approvers.items():
        if not approvers:
            logging.warning(f"No approvers assigned to tool '{tool_func.__name__}'. Skipping registration.")
            continue

        # Add the tool and its approvers to the supervision context
        supervision_context.add_supervised_function(
            func=tool_func,
            supervision_functions=[approvers],
            ignored_attributes=[]
        )

    # Register the tools and supervisors with the Sentinel API
    register_tools_and_supervisors(run_id=run_id)
   
   
def _register_inspect_approvals_from_approval_file(run_id: UUID, approval_file: str):
    """
    Reads the inspect approval YAML file and registers the approvals and tools for the run.
    
    Args:
        run_id (UUID): The ID of the run.
        approval_file (str): The path to the inspect approval YAML file.
    """
    from entropy_labs.supervision.config import supervision_config 
    
    # Read the approval file
    with open(approval_file, 'r') as file:
        approvals = yaml.safe_load(file)

    client = supervision_config.client
    if client is None:
        raise Exception("Client not set in the supervision config. Please set the client before calling this function.")

    run = supervision_config.get_run_by_id(run_id)
    if run is None:
        raise Exception(f"Run with ID {run_id} not found in supervision config.")
    supervision_context = run.supervision_context

    supervised_tools = {}
    # For each approver in the approval file
    for approver in approvals.get('approvers', []):
        # Get the tools pattern (may be a wildcard)
        tools_pattern = approver.get('tools', '*')

        # Find tools from the registry matching the pattern
        tools = registry_find(lambda x: x.type == "tool")

        # Filter tools based on tools_pattern
        matching_tools = []
        for tool in tools:
            tool_name = tool.__registry_info__.name.removeprefix('inspect_ai/')
            # tools_pattern may be a list of patterns
            if isinstance(tools_pattern, list):
                if any(fnmatch.fnmatch(tool_name, pattern) for pattern in tools_pattern):
                    matching_tools.append(tool)
            else:
                if fnmatch.fnmatch(tool_name, tools_pattern):
                    matching_tools.append(tool)

        # For each matching tool, add supervised function to the supervision context
        for tool in matching_tools:
            # The tool is a function decorated with @tool in inspect_ai
            func = tool

            # Get the supervisor function (approval function)
            approver_name = approver.get('name')
            approval_funcs = registry_find(lambda x: x.type == "approver" and x.name == approver_name)

            if approver_name =='auto':
                logging.info(f"Auto approval function '{approver_name}' found in the registry. Not registering approval function.")
                continue

            if not approval_funcs:
                logging.warning(f"Approval function '{approver_name}' not found in the registry.")
                continue

            approval_func = approval_funcs[0]

            # Configure the approval function with attributes from the approval file
            supervisor_attributes = {k: v for k, v in approver.items() if k not in ['name', 'tools']}
            approval_func_initialised = approval_func(**supervisor_attributes)

            if func not in supervised_tools:
                supervised_tools[func] = [approval_func_initialised]
            else:
                supervised_tools[func].append(approval_func_initialised)

    for tool_func in supervised_tools:
        supervision_functions = supervised_tools[tool_func]
        supervision_context.add_supervised_function(
            func=tool_func,
            supervision_functions=[supervision_functions],
            ignored_attributes=[]
        )

    # Register the tools and supervisors with the Sentinel API
    register_tools_and_supervisors(run_id=run_id)


def register_inspect_samples_with_entropy_labs(tasks: Task, project_id: UUID, approval: str) -> list[Sample]:
    """
    Registers samples with Entropy Labs by creating a project, task, and run.
    Maps tools to their respective approvers and registers the approvals.
    """
    samples = []
    for idx, sample in enumerate(tasks.dataset.samples):
        # We need to assign an ID to each sample and register the task
        if sample.id is None:
            print(f"Each sample must have an ID, adding {idx} to the ID")
            sample.id = f"{idx}"
        task_id = register_task(project_id=project_id, task_name=sample.id)
        run_id = create_run(project_id=project_id, task_id=task_id, run_name=sample.id)
        _register_inspect_approvals_from_approval_file(run_id=run_id, approval_file=approval)
        samples.append(sample)
    return samples    


@solver
def register_inspect_samples_with_entropy_labs_solver(project_name: str, entropy_labs_backend_url: str) -> Solver:
    """
    Registers samples with Entropy Labs by creating a project, task, and run.
    Maps tools to their respective approvers and registers the approvals.

    Args:
        project_name (str): The name of the project to register.
        entropy_labs_backend_url (str): The backend URL for Entropy Labs.

    Returns:
        Solver: An asynchronous solver function that processes the task state.
    """
    async def solve(state: TaskState, generate) -> TaskState:
        # Register project, task, and run
        sample_id = str(state.sample_id)
        project_id = register_project(project_name=project_name, entropy_labs_backend_url=entropy_labs_backend_url)
        task_id = register_task(project_id=project_id, task_name=sample_id)
        run_id = create_run(project_id=project_id, task_id=task_id, run_name=sample_id)
        
        # Extract tool approvers
        from inspect_ai.approval._apply import _tool_approver
        approver = _tool_approver.get(None)
        tool_approvers = _extract_tool_approvers(approver)
        
        # Find all registered tools
        tools = registry_find(lambda x: x.type == "tool")
        
        # Map tools to their approvers
        tool_to_approvers = {}
        for tool in tools:
            # Create a dummy tool call for this tool
            dummy_call = ToolCall(
                id="dummy-id",
                function=tool.__qualname__,
                arguments={},
                type="function"
            )
            
            # Get all approvers for this tool
            matching_approvers = list(tool_approvers(dummy_call))
            tool_to_approvers[tool] = matching_approvers
        
        # Register the tool-approver mappings
        _register_inspect_approvals(run_id=run_id, tool_to_approvers=tool_to_approvers)
        
        return state

    return solve



@scorer(metrics=[accuracy(), stderr()])
def entropy_labs_web_ui_scorer(timeout: Optional[int] = 300, wait_for_result: bool = True) -> Scorer:
    async def score(state: TaskState, target: Target) -> Score:
        """
        A scorer function for evaluating the results of a task run in the Entropy Labs UI.

        Args:
            state (TaskState): The current state of the task being evaluated.
            target (Target): The target to evaluate against.

        Returns:
            Score: A Score object containing the evaluation result.
        """
        # Update the run status to 'completed' in the Entropy Labs system
        update_run_status_by_sample_id(str((state.sample_id)), status="completed")
        
        # Retrieve the result of the run if waiting for the result is enabled
        if wait_for_result:
            print(f"Waiting for human to score sample in the web UI {state.sample_id}")
            result = get_sample_result(str(state.sample_id), timeout=timeout)
        else:
            result = ''
        
        # Return a Score object with a value of 1 if the result is 'passed', otherwise 0
        return Score(
            value=1 if result == 'passed' else 0,
            answer="",
            explanation="",
        )
    return score


import datetime
from typing import Any, Dict, List, Type, TypeVar, cast
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

T = TypeVar("T", bound="Project")


@_attrs_define
class Project:
    """
    Attributes:
        id (UUID):
        name (str):
        created_at (datetime.datetime):
        run_result_tags (List[str]):
    """

    id: UUID
    name: str
    created_at: datetime.datetime
    run_result_tags: List[str]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = str(self.id)

        name = self.name

        created_at = self.created_at.isoformat()

        run_result_tags = self.run_result_tags

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "name": name,
                "created_at": created_at,
                "run_result_tags": run_result_tags,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = UUID(d.pop("id"))

        name = d.pop("name")

        created_at = isoparse(d.pop("created_at"))

        run_result_tags = cast(List[str], d.pop("run_result_tags"))

        project = cls(
            id=id,
            name=name,
            created_at=created_at,
            run_result_tags=run_result_tags,
        )

        project.additional_properties = d
        return project

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.message_role import MessageRole
from ..models.message_type import MessageType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.tool_call import ToolCall


T = TypeVar("T", bound="Message")


@_attrs_define
class Message:
    """
    Attributes:
        role (MessageRole):
        content (str):
        id (Union[Unset, UUID]):
        type (Union[Unset, MessageType]):
        source (Union[Unset, str]):
        tool_calls (Union[Unset, List['ToolCall']]):
    """

    role: MessageRole
    content: str
    id: Union[Unset, UUID] = UNSET
    type: Union[Unset, MessageType] = UNSET
    source: Union[Unset, str] = UNSET
    tool_calls: Union[Unset, List["ToolCall"]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        role = self.role.value

        content = self.content

        id: Union[Unset, str] = UNSET
        if not isinstance(self.id, Unset):
            id = str(self.id)

        type: Union[Unset, str] = UNSET
        if not isinstance(self.type, Unset):
            type = self.type.value

        source = self.source

        tool_calls: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.tool_calls, Unset):
            tool_calls = []
            for tool_calls_item_data in self.tool_calls:
                tool_calls_item = tool_calls_item_data.to_dict()
                tool_calls.append(tool_calls_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "role": role,
                "content": content,
            }
        )
        if id is not UNSET:
            field_dict["id"] = id
        if type is not UNSET:
            field_dict["type"] = type
        if source is not UNSET:
            field_dict["source"] = source
        if tool_calls is not UNSET:
            field_dict["tool_calls"] = tool_calls

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.tool_call import ToolCall

        d = src_dict.copy()
        role = MessageRole(d.pop("role"))

        content = d.pop("content")

        _id = d.pop("id", UNSET)
        id: Union[Unset, UUID]
        if isinstance(_id, Unset):
            id = UNSET
        else:
            id = UUID(_id)

        _type = d.pop("type", UNSET)
        type: Union[Unset, MessageType]
        if isinstance(_type, Unset):
            type = UNSET
        else:
            type = MessageType(_type)

        source = d.pop("source", UNSET)

        tool_calls = []
        _tool_calls = d.pop("tool_calls", UNSET)
        for tool_calls_item_data in _tool_calls or []:
            tool_calls_item = ToolCall.from_dict(tool_calls_item_data)

            tool_calls.append(tool_calls_item)

        message = cls(
            role=role,
            content=content,
            id=id,
            type=type,
            source=source,
            tool_calls=tool_calls,
        )

        message.additional_properties = d
        return message

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, List, Dict
import uuid
from datetime import datetime
from inspect_ai.solver import TaskState
from inspect_ai.model import ChatMessage, ChatMessageAssistant, ChatMessageSystem, ChatMessageTool, ChatMessageUser, ModelName
from inspect_ai.tool import ToolCall
from inspect_ai.tool._tool_call import ToolCallError
import json

def extract_messages_from_events(events: List[Dict[str, Any]]) -> List[ChatMessage]:
    """
    Extract messages from LangChain events.

    Args:
        events (List[Dict[str, Any]]): The list of LangChain events.

    Returns:
        List[ChatMessage]: A list of ChatMessage objects representing the conversation.
    """
    messages = []
    for event in events:
        if event['event'] == 'on_chat_model_start':
            # Extract messages from the start event
            message_batches = event['data'].get('messages', [])
            for batch in message_batches:
                for msg_dict in batch:
                    msg_type = msg_dict.get('type')
                    if msg_type != 'human':
                        continue  # Skip non-human messages
                    content = msg_dict.get('data', {}).get('content', '').strip()
                    if not content:
                        continue  # Skip empty messages
                    
                    # Check for duplicate last user message
                    if messages:
                        last_user_messages = [msg for msg in messages if msg.role == 'user']
                        if len(last_user_messages) > 0 and last_user_messages[-1].content == content:
                            continue  # Skip adding duplicate user message

                    messages.append(
                        ChatMessageUser(
                            role='user',
                            content=content,
                            source='input'
                        )
                    )
        elif event['event'] == 'on_llm_end':
            # Assistant response
            response = event['data'].get('response', {})
            generations = response.get('generations', [])
            for generation_list in generations:
                for generation in generation_list:
                    message = generation.get('message', {})
                    content = message.get('content', '')
                    additional_kwargs = message.get('additional_kwargs', {})
                    tool_calls = additional_kwargs.get('tool_calls', None)
                    if tool_calls:
                        tool_calls_list = []
                        # Assistant made a function call
                        for tool_call in tool_calls:
                            tool_call_id = tool_call.get('id', '')
                            function = tool_call.get('function', {})
                            function_name = function.get('name')
                            arguments = function.get('arguments', '')
                            # parse arguments to json
                            try:
                                arguments = json.loads(arguments)
                            except json.JSONDecodeError:
                                arguments = {}
                            tool_call_type = tool_call.get('type', '')
                            tool_call = ToolCall(
                                id=tool_call_id,
                                function=function_name,
                                arguments=arguments,
                                type=tool_call_type
                            )
                            tool_calls_list.append(tool_call)
                        messages.append(
                            ChatMessageAssistant(
                                role='assistant',
                                content=content,
                                tool_calls=tool_calls_list,
                                source='generate'
                            )
                        )
                    else:
                        # Regular assistant message
                        messages.append(
                            ChatMessageAssistant(
                                role='assistant',
                                content=content,
                                tool_calls=[],
                                source='generate'
                            )
                        )
        elif event['event'] == 'on_tool_end':
            # Tool response
            output = event['data'].get('output', {})
            content = output.get('content', '')
            tool_call_id = output.get('tool_call_id', '')
            function_name = output.get('name', '')
            error_data = output.get('error', None)
            error = None
            if error_data:
                error = ToolCallError(
                    type=error_data.get('type', ''),
                    message=error_data.get('message', '')
                )
            messages.append(
                ChatMessageTool(
                    role='tool',
                    content=content,
                    tool_call_id=tool_call_id,
                    function=function_name,
                    error=error,
                    source='generate'
                )
            )
        elif event['event'] == 'on_tool_error':
            # Tool error response
            error_message = event['data'].get('error', '')
            run_id = event.get('run_id', '')
            # You might want to map run_id to tool_call_id and function_name if available
            messages.append(
                ChatMessageTool(
                    role='tool',
                    content='',
                    tool_call_id=run_id,  # Placeholder, update if you have mapping
                    function=None,        # Placeholder, update if you have mapping
                    error=ToolCallError(
                        type='ToolError',
                        message=error_message
                    ),
                    source='generate'
                )
            )
    return messages





import json
from typing import Any, Callable
from openai import OpenAI
from pydantic import BaseModel
from .config import PREFERRED_LLM_MODEL

client = OpenAI()

def create_schema_from_function(func: Callable) -> dict:
    """Create a JSON schema from a function's return type annotation."""
    return_type = func.__annotations__.get('return')
    if not return_type or not issubclass(return_type, BaseModel):
        raise ValueError("Function must have a Pydantic BaseModel return type annotation")
    
    return return_type.model_json_schema()

def sample_from_llm(func: Callable, *args, **kwargs) -> Any:
    """Sample a response from the LLM based on the function's signature and return type."""
    func_name = func.__name__
    func_doc = func.__doc__ or "No description available."
    schema = create_schema_from_function(func)
    
    # Prepare messages for the chat completion
    messages = [
        {
            "role": "system",
            "content": (
                f"You are simulating the API function '{func_name}'. {func_doc} "
                f"Provide a response that matches the following JSON schema: {json.dumps(schema)}"
            ),
        },
        {
            "role": "user",
            "content": (
                f"Please generate a valid response for the function '{func_name}' "
                f"with the following arguments: args={args}, kwargs={kwargs}"
            ),
        },
    ]

    try:
        completion = client.chat.completions.create(
            model=PREFERRED_LLM_MODEL,
            messages=messages,
            # Using functions and function_call as per the API
            functions=[
                {
                    "name": func_name,
                    "description": func_doc,
                    "parameters": schema,
                }
            ],
            function_call={"name": func_name},
            temperature=0,
        )
        
        # Parse the function call arguments from the response
        response = completion.choices[0].message.function_call.arguments
        response_data = json.loads(response)
        
        # Validate the response against the schema
        return_type = func.__annotations__['return']
        validated_response = return_type(**response_data)
        
        return validated_response
    except Exception as e:
        print(f"Error sampling from LLM: {str(e)}")
        raise ValueError(
            f"Failed to generate a valid response from LLM for function '{func_name}'"
        )



import asyncio
from typing import Any, Callable, List, Optional
from functools import wraps

from .config import supervision_config, SupervisionDecisionType, SupervisionDecision
from ..mocking.policies import MockPolicy
from ..utils.utils import create_random_value
from .llm_sampling import sample_from_llm
import random
from uuid import UUID, uuid4
from entropy_labs.sentinel_api_client.sentinel_api_client.models.supervisor_type import SupervisorType
from entropy_labs.sentinel_api_client.sentinel_api_client.models.tool_request import ToolRequest
from entropy_labs.sentinel_api_client.sentinel_api_client.models.arguments import Arguments
from entropy_labs.sentinel_api_client.sentinel_api_client.models.message import Message
from entropy_labs.sentinel_api_client.sentinel_api_client.models.task_state import TaskState
from entropy_labs.sentinel_api_client.sentinel_api_client.models.tool_request_group import ToolRequestGroup

def supervise(
    mock_policy: Optional[MockPolicy] = None,
    mock_responses: Optional[List[Any]] = None,
    supervision_functions: Optional[List[List[Callable]]] = None,
    ignored_attributes: Optional[List[str]] = None
):
    """
    Decorator that supervises a function.
    
    Args:
        mock_policy (Optional[MockPolicy]): Mock policy to use. Defaults to None.
        mock_responses (Optional[List[Any]]): Mock responses to use. Defaults to None.
        supervision_functions (Optional[List[List[Callable]]]): Supervision functions to use. Defaults to None.
        ignored_attributes (Optional[List[str]]): Ignored attributes. Defaults to None.
    """
    if supervision_functions and len(supervision_functions) == 1 and isinstance(supervision_functions[0], list):
        supervision_functions = [supervision_functions[0]]

    def decorator(func):
        # Register the supervised function in SupervisionConfig's pending functions
        supervision_config.register_pending_supervised_function(
            func, supervision_functions, ignored_attributes
        )

        @wraps(func)
        def wrapper(*tool_args, **tool_kwargs):
            
            from entropy_labs.api.sentinel_api_client_helper import create_tool_request_group, get_supervisor_chains_for_tool, send_supervision_request, send_supervision_result, _serialize_arguments

            supervision_context = supervision_config.get_all_runs()[0].supervision_context
            client = supervision_config.client  # Get the Sentinel API client
            # TODO: This for now assumes there is only one run
            
            print(f"\n--- Supervision ---")
            print(f"Function Name: {func.__qualname__}")
            print(f"Description: {func.__doc__}")
            print(f"Arguments: {tool_args}, {tool_kwargs.keys()}")

            # Retrieve the tool_id from the registry
            entry = supervision_context.get_supervised_function_entry(func.__qualname__)
            if entry:
                tool_id = entry['tool_id']
                ignored_attributes = entry['ignored_attributes']
                run_id = entry['run_id']
            else:
                raise Exception(f"Tool ID for function {func.__name__} not found in the registry.")
            
            # TODO if n > 1, it would be handled here
            arguments_dict = _serialize_arguments(tool_args, tool_kwargs)
            
            tool_requests = [ToolRequest(tool_id=tool_id, 
                                         message=supervision_context.get_api_messages()[-1],
                                         arguments=Arguments.from_dict(arguments_dict), 
                                         task_state=supervision_context.to_task_state())]
            tool_request_group = create_tool_request_group(tool_id, tool_requests, client)
            tool_request = tool_request_group.tool_requests[0] #TODO: Fix for n > 1
            supervisors_chains = get_supervisor_chains_for_tool(tool_id, client)

            # Determine effective mock policy
            effective_mock_policy = (
                supervision_config.global_mock_policy
                if supervision_config.override_local_policy
                else (mock_policy or supervision_config.global_mock_policy)
            )
            # Handle mocking
            if effective_mock_policy != MockPolicy.NO_MOCK:
                mock_result = handle_mocking(func, effective_mock_policy, mock_responses, *tool_args, **tool_kwargs)
                print(f"Mocking function execution. Mock result: {mock_result}")
                return mock_result
            # Supervision logic
            if not supervisors_chains:
                print(f"No supervisors found for function {func.__name__}. Executing function.")
                return func(*tool_args, **tool_kwargs)
            all_decisions = []
            for supervisor_chain in supervisors_chains:
                chain_decisions = []
                # We send supervision request to the API
                supervisors = supervisor_chain.supervisors
                supervisor_chain_id = supervisor_chain.chain_id
                for position_in_chain, supervisor in enumerate(supervisors):
                    supervision_request_id = send_supervision_request(supervisor_chain_id=supervisor_chain_id, supervisor_id=supervisor.id, request_group_id=tool_request_group.id, position_in_chain=position_in_chain)

                    decision = None
                    supervisor_func = supervision_context.get_supervisor_by_id(supervisor.id)
                    if supervisor_func is None:
                        print(f"No local supervisor function found for ID {supervisor.id}. Skipping.")
                        return None  # Continue to next supervisor

                    # Execute supervisor function
                    decision = call_supervisor_function(supervisor_func, func, supervision_context, supervision_request_id=supervision_request_id, ignored_attributes=ignored_attributes, tool_args=tool_args, tool_kwargs=tool_kwargs, decision=decision)
                    chain_decisions.append(decision)
                    print(f"Supervisor decision: {decision.decision}")

                    if supervisor.type != SupervisorType.HUMAN_SUPERVISOR:
                        # We send the decision to the API
                        send_supervision_result(
                            supervision_request_id=supervision_request_id,
                            request_group_id=tool_request_group.id,
                            tool_id=tool_id,
                            supervisor_id=supervisor.id,
                            decision=decision,
                            client=client,
                            tool_request=tool_request, #TODO: Fix for n > 1
                            tool_args=tool_args, #TODO: If modified, send modified args and kwargs
                            tool_kwargs=tool_kwargs
                    )
                    # Handle the decision
                    if decision.decision == SupervisionDecisionType.APPROVE:
                        all_decisions.append(decision)
                        break
                    elif decision.decision == SupervisionDecisionType.REJECT:
                        explanations = " ".join([f"Decision: {d.decision}, Explanation: {d.explanation}" for d in chain_decisions])
                        return (f"The agent requested to execute {func.__qualname__} but it was rejected. "
                                f"Explanation: {decision.explanation}. "
                                f"Chain Explanations: {explanations}\n"
                                "This is not a message from the user but from a supervisor system that is helping the agent to improve its behavior. You should try different action using the feedback!")
                    elif decision.decision == SupervisionDecisionType.ESCALATE:
                        # if last decision in chain is escalate, we continue to the next supervisor chain
                        if position_in_chain == len(supervisors) - 1:
                            all_decisions.append(decision)
                        continue
                    elif decision.decision == SupervisionDecisionType.MODIFY:
                        all_decisions.append(decision)
                        break
                    elif decision.decision == SupervisionDecisionType.TERMINATE:
                        explanations = " ".join([f"Decision: {d.decision}, Explanation: {d.explanation}" for d in chain_decisions])
                        return (f"Execution of {func.__qualname__} should be terminated. "
                                f"Explanation: {decision.explanation}. "
                                f"Chain Explanations: {explanations}\n"
                                "This is not a message from the user but from a supervisor system that is helping the agent to improve its behavior. You should try different action using the feedback!")
                    else:
                        print(f"Unknown decision: {decision.decision}. Cancelling execution.")
                        explanations = " ".join([f"Decision: {d.decision}, Explanation: {d.explanation}" for d in chain_decisions])
                        return (f"Execution of {func.__qualname__} was cancelled due to an unknown supervision decision. "
                                f"Chain Explanations: {explanations}\n"
                                "This is not a message from the user but from a supervisor system that is helping the agent to improve its behavior. You should try different action using the feedback!")

            # Check decisions and apply modifications if any
            if all(decision.decision in [SupervisionDecisionType.APPROVE, SupervisionDecisionType.MODIFY] for decision in all_decisions):
                print("All decisions approved or modified.")
                # Start with original arguments
                final_args = list(tool_args)
                final_kwargs = tool_kwargs.copy()

                # Apply modifications while respecting ignored_attributes
                for decision in all_decisions:
                    if decision.decision == SupervisionDecisionType.MODIFY and decision.modified:
                        #TODO: Make sure this works correctly
                        tool_args = decision.modified.tool_args
                        # Update keyword arguments
                        if decision.modified.tool_kwargs:
                            for key, value in decision.modified.tool_kwargs.items():
                                if ignored_attributes and key in ignored_attributes:
                                    continue
                                final_kwargs[key] = value
                        break

                # Call the function with modified arguments
                return func(*final_args, **final_kwargs)
            else:
                explanations = " ".join([f"Supervisor {idx}: Decision: {d.decision}, Explanation: {d.explanation} \n" for idx, d in enumerate(all_decisions)])
                return (f"The agent requested to execute a function but it was rejected by some supervisors.\n"
                        f"Chain Explanations: \n{explanations}\n"
                        "This is not a message from the user but from a supervisor system that is helping the agent to improve its behavior. You should try something else!")
        return wrapper
    return decorator


def call_supervisor_function(supervisor_func, func, supervision_context, supervision_request_id: UUID, ignored_attributes: List[str], tool_args: List[Any], tool_kwargs: dict[str, Any], decision: Optional[SupervisionDecision] = None):
    if asyncio.iscoroutinefunction(supervisor_func):
        decision = asyncio.run(supervisor_func(
            func, supervision_context=supervision_context, supervision_request_id=supervision_request_id, ignored_attributes=ignored_attributes, tool_args=tool_args, tool_kwargs=tool_kwargs, decision=decision
        ))
    else:
        decision = supervisor_func(
            func, supervision_context=supervision_context, supervision_request_id=supervision_request_id, ignored_attributes=ignored_attributes, tool_args=tool_args, tool_kwargs=tool_kwargs, decision=decision
        )
    return decision

def handle_mocking(func, mock_policy, mock_responses, *args, **kwargs):
    """Handle different mock policies."""
    if mock_policy == MockPolicy.NO_MOCK:
        return func(*args, **kwargs)
    elif mock_policy == MockPolicy.SAMPLE_LIST:
        if mock_responses:
            return random.choice(mock_responses)
        else:
            raise ValueError("No mock responses provided for SAMPLE_LIST policy")
    elif mock_policy == MockPolicy.SAMPLE_RANDOM:
        tool_return_type = func.__annotations__.get('return', None)
        if tool_return_type:
            return create_random_value(tool_return_type)
        else:
            raise ValueError("No return type specified for the function")
    elif mock_policy == MockPolicy.SAMPLE_PREVIOUS_CALLS:
        try:
            return supervision_config.get_mock_response(func.__name__)  # TODO: Make sure this works
        except ValueError as e:
            print(f"Warning: {str(e)}. Falling back to actual function execution.")
            return func(*args, **kwargs)
    elif mock_policy == MockPolicy.SAMPLE_LLM:
        return sample_from_llm(func, *args, **kwargs)
    else:
        raise ValueError(f"Unsupported mock policy: {mock_policy}")



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.choice import Choice
    from ..models.usage import Usage


T = TypeVar("T", bound="Output")


@_attrs_define
class Output:
    """
    Attributes:
        model (Union[Unset, str]):
        choices (Union[Unset, List['Choice']]):
        usage (Union[Unset, Usage]):
    """

    model: Union[Unset, str] = UNSET
    choices: Union[Unset, List["Choice"]] = UNSET
    usage: Union[Unset, "Usage"] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        model = self.model

        choices: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.choices, Unset):
            choices = []
            for choices_item_data in self.choices:
                choices_item = choices_item_data.to_dict()
                choices.append(choices_item)

        usage: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.usage, Unset):
            usage = self.usage.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if model is not UNSET:
            field_dict["model"] = model
        if choices is not UNSET:
            field_dict["choices"] = choices
        if usage is not UNSET:
            field_dict["usage"] = usage

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.choice import Choice
        from ..models.usage import Usage

        d = src_dict.copy()
        model = d.pop("model", UNSET)

        choices = []
        _choices = d.pop("choices", UNSET)
        for choices_item_data in _choices or []:
            choices_item = Choice.from_dict(choices_item_data)

            choices.append(choices_item)

        _usage = d.pop("usage", UNSET)
        usage: Union[Unset, Usage]
        if isinstance(_usage, Unset):
            usage = UNSET
        else:
            usage = Usage.from_dict(_usage)

        output = cls(
            model=model,
            choices=choices,
            usage=usage,
        )

        output.additional_properties = d
        return output

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Callable, Dict, List, Optional, Set, Any
from entropy_labs.supervision.config import SupervisionDecisionType, SupervisionDecision, supervision_config, SupervisionContext
from inspect_ai.tool import ToolCall
from entropy_labs.sentinel_api_client.sentinel_api_client.client import Client
from uuid import UUID

def human_supervisor(agent_id: str = "default_agent", timeout: int = 300, n: int = 1):
    async def supervisor(func: Callable, supervision_context: SupervisionContext, supervision_request_id: Optional[UUID] = None, ignored_attributes: List[str] = [], 
                         tool_args: List[Any] = [], tool_kwargs: dict[str, Any] = {}, decision: Optional[SupervisionDecision] = None) -> SupervisionDecision:
        """
        Human supervisor that requests approval via backend API or CLI.
        """
        from entropy_labs.supervision.common import human_supervisor_wrapper
        
        context = supervision_config.context
        
        # TODO: Reuse Supervise Protocol

        # Create TaskState from context - Right now we are using Inspect AI TaskState format to send to the backend, we need to transform langchain format to TaskState format
        task_state = context.to_task_state()
        id = 'tool_id'
        tool_call = ToolCall(id=id,function=func.__name__, arguments=tool_kwargs, type='function')
        client = supervision_config.client
        
        supervisor_decision = human_supervisor_wrapper(task_state=task_state, call=tool_call, timeout=timeout, use_inspect_ai=False, n=n, supervision_request_id=supervision_request_id, client=client)

        return supervisor_decision

    supervisor.__name__ = human_supervisor.__name__
    supervisor.supervisor_attributes = {"timeout": timeout, "n": n}
    return supervisor

def bash_supervisor(
    allowed_commands: List[str],
    allow_sudo: bool = False,
    command_specific_rules: Optional[Dict[str, List[str]]] = None
):
    """
    Decorator that creates a supervisor function to check bash commands.
    
    Args:
        allowed_commands (List[str]): List of allowed bash commands.
        allow_sudo (bool): Whether to allow sudo commands. Defaults to False.
        command_specific_rules (Optional[Dict[str, List[str]]]): Dictionary of command-specific rules.
    
    Returns:
        Callable: A supervisor function that checks bash commands.
    """
    def supervisor(func: Callable, ignored_attributes: List[str] = [], tool_kwargs: dict[str, Any] = {}) -> SupervisionDecision:
        from entropy_labs.supervision.common import check_bash_command
        
        command = tool_kwargs.get('command', '')
        is_approved, explanation = check_bash_command(
            command, allowed_commands, allow_sudo, command_specific_rules
        )
        if is_approved:
            return SupervisionDecision(
                decision=SupervisionDecisionType.APPROVE,
                explanation=explanation
            )
        else:
            return SupervisionDecision(
                decision=SupervisionDecisionType.REJECT,
                explanation=explanation
            )
    
    supervisor.__name__ = bash_supervisor.__name__
    supervisor.supervisor_attributes = {"allowed_commands": allowed_commands, "allow_sudo": allow_sudo, "command_specific_rules": command_specific_rules}
    return supervisor

def python_supervisor(
    allowed_modules: List[str],
    allowed_functions: List[str],
    disallowed_builtins: Optional[Set[str]] = None,
    sensitive_modules: Optional[Set[str]] = None,
    allow_system_state_modification: bool = False
):
    """
    Decorator that creates a supervisor function to check Python code.
    
    Args:
        allowed_modules (List[str]): List of allowed Python modules.
        allowed_functions (List[str]): List of allowed functions.
        disallowed_builtins (Optional[Set[str]]): Set of disallowed built-in functions.
        sensitive_modules (Optional[Set[str]]): Set of sensitive modules to be blocked.
        allow_system_state_modification (bool): Whether to allow modification of system state.
    
    Returns:
        Callable: A supervisor function that checks Python code.
    """
    def supervisor(func: Callable, ignored_attributes: List[str] = [], tool_kwargs: dict[str, Any] = {}) -> SupervisionDecision:
        from entropy_labs.supervision.common import check_python_code
        
        code = tool_kwargs.get('code', '')
        is_approved, explanation = check_python_code(
            code, allowed_modules, allowed_functions, disallowed_builtins,
            sensitive_modules, allow_system_state_modification
        )
        if is_approved:
            return SupervisionDecision(
                decision=SupervisionDecisionType.APPROVE,
                explanation=explanation
            )
        else:
            return SupervisionDecision(
                decision=SupervisionDecisionType.REJECT,
                explanation=explanation
            )
    
    supervisor.__name__ = python_supervisor.__name__
    supervisor.supervisor_attributes = {"allowed_modules": allowed_modules, "allowed_functions": allowed_functions, "disallowed_builtins": disallowed_builtins, "sensitive_modules": sensitive_modules, "allow_system_state_modification": allow_system_state_modification}
    return supervisor
